/*
 * layout.c --
 *
 * Generic layout processing.
 *
 * Copyright (c) 2003 Joe English.  Freely redistributable.
 *
 * layout.c,v 1.38 2004/04/06 02:19:46 jenglish Exp
 */

#include <string.h>
#include <tk.h>
#include "tkTheme.h"

#define MAX(a,b) (a > b ? a : b)
#define MIN(a,b) (a < b ? a : b)

/*
 * +++ TTK_Box and TTK_Padding utilities:
 */

TTK_Box TTK_MakeBox(int x, int y, int width, int height)
{
    TTK_Box b;
    b.x = x; b.y = y; b.width = width; b.height = height;
    return b;
}

int TTK_BoxContains(TTK_Box box, int x, int y)
{
    return box.x <= x && x < box.x + box.width
	&& box.y <= y && y < box.y + box.height;
}

/*
 * packTop, packBottom, packLeft, packRight --
 * 	Carve out a parcel of the specified height (resp width) 
 * 	from the specified cavity.
 *
 * Returns: 
 * 	The new parcel.
 *
 * Side effects:
 * 	Adjust the cavity.
 */

static TTK_Box packTop(TTK_Box *cavity, int height)
{
    TTK_Box parcel;
    height = MIN(height, cavity->height);
    parcel = TTK_MakeBox(cavity->x, cavity->y, cavity->width, height);
    cavity->y += height;
    cavity->height -= height;
    return parcel;
}

static TTK_Box packBottom(TTK_Box *cavity, int height)
{
    height = MIN(height, cavity->height);
    cavity->height -= height;
    return TTK_MakeBox(
	cavity->x, cavity->y + cavity->height,
	cavity->width, height);
}

static TTK_Box packLeft(TTK_Box *cavity, int width)
{
    TTK_Box parcel;
    width = MIN(width, cavity->width);
    parcel = TTK_MakeBox(cavity->x, cavity->y, width,cavity->height);
    cavity->x += width;
    cavity->width -= width;
    return parcel;
}

static TTK_Box packRight(TTK_Box *cavity, int width)
{
    width = MIN(width, cavity->width);
    cavity->width -= width;
    return TTK_MakeBox(cavity->x + cavity->width, 
	    cavity->y, width, cavity->height);
}

/*
 * TTK_PackBox --
 * 	Carve out a parcel of the specified size on the specified side
 * 	in the specified cavity.
 *
 * Returns: 
 * 	The new parcel.
 *
 * Side effects:
 * 	Adjust the cavity.
 */

TTK_Box TTK_PackBox(TTK_Box *cavity, int width, int height, TTK_Side side)
{
    switch (side) {
	default:
	case TTK_SIDE_TOP:	return packTop(cavity, height);
	case TTK_SIDE_BOTTOM:	return packBottom(cavity, height);
	case TTK_SIDE_LEFT:	return packLeft(cavity, width);
	case TTK_SIDE_RIGHT:	return packRight(cavity, width);
    }
}

/*
 * TTK_PadBox --
 * 	Shrink a box by the specified padding amount.
 */
TTK_Box TTK_PadBox(TTK_Box b, TTK_Padding p)
{
    b.x += p.left;
    b.y += p.top;
    b.width -= (p.left + p.right);
    b.height -= (p.top + p.bottom);
    return b;
}

/*
 * TTK_StickBox --
 * 	Place a box of size w * h in the specified parcel,
 * 	according to the specified sticky bits.
 */
TTK_Box TTK_StickBox(TTK_Box parcel, int width, int height, unsigned sticky)
{
    int dx, dy;

    if (width > parcel.width) width = parcel.width;
    if (height > parcel.height) height = parcel.height;

    dx = parcel.width - width;
    dy = parcel.height - height;

    /*
     * X coordinate adjustment:
     */
    switch (sticky & (TTK_STICK_W | TTK_STICK_E)) 
    {
	case TTK_STICK_W | TTK_STICK_E:
	    /* no-op -- use entire parcel width */
	    break;
	case TTK_STICK_W:
	    parcel.width = width;
	    break;
	case TTK_STICK_E:
	    parcel.x += dx;
	    parcel.width = width;
	    break;
	default :
	    parcel.x += dx / 2;
	    parcel.width = width;
	    break;
    }

    /*
     * Y coordinate adjustment:
     */
    switch (sticky & (TTK_STICK_N | TTK_STICK_S)) 
    {
	case TTK_STICK_N | TTK_STICK_S:
	    /* use entire parcel height */
	    break;
	case TTK_STICK_N:
	    parcel.height = height;
	    break;
	case TTK_STICK_S:
	    parcel.y += dy;
	    parcel.height = height;
	    break;
	default :
	    parcel.y += dy / 2;
	    parcel.height = height;
	    break;
    }

    return parcel;
}

/*
 * TTK_AnchorBox --
 * 	Place a box of size w * h in the specified parcel,
 * 	according to the specified anchor.
 */

static unsigned AnchorToSticky(int anchor)
{
    switch (anchor)
    {
	case TK_ANCHOR_N:	return TTK_STICK_N;
	case TK_ANCHOR_NE:	return TTK_STICK_N | TTK_STICK_E;
	case TK_ANCHOR_E:	return TTK_STICK_E;
	case TK_ANCHOR_SE:	return TTK_STICK_S | TTK_STICK_E;
	case TK_ANCHOR_S:	return TTK_STICK_S;
	case TK_ANCHOR_SW:	return TTK_STICK_S | TTK_STICK_W;
	case TK_ANCHOR_W:	return TTK_STICK_W;
	case TK_ANCHOR_NW:	return TTK_STICK_N | TTK_STICK_W;
	default:
	case TK_ANCHOR_CENTER:	return 0;
    }
}

TTK_Box TTK_AnchorBox(TTK_Box parcel, int width, int height, int anchor)
{
    return TTK_StickBox(parcel, width, height, AnchorToSticky(anchor));
}

/*
 * TTK_GetPaddingFromObj --
 *
 * 	Extract a padding specification from a Tcl_Obj * scaled
 * 	to work with a particular Tk_Window.
 *
 * 	The string representation of a TTK_Padding is a list
 * 	of one to four Tk_Pixel specifications, corresponding
 * 	to the left, top, right, and bottom padding.
 *
 * 	If the 'bottom' (fourth) element is missing, it defaults to 'top'.
 * 	If the 'right' (third) element is missing, it defaults to 'left'.
 * 	If the 'top' (second) element is missing, it defaults to 'left'.
 *
 * 	The internal representation is a Tcl_ListObj containing
 * 	one to four Tk_PixelObj objects.
 *
 * Returns:
 * 	TCL_OK or TCL_ERROR.  In the latter case an error message is
 * 	left in 'interp' and '*paddingPtr' is set to all-zeros.
 * 	Otherwise, *paddingPtr is filled in with the padding specification.
 *
 */
int TTK_GetPaddingFromObj(
    Tcl_Interp *interp,
    Tk_Window tkwin,
    Tcl_Obj *objPtr,
    TTK_Padding *pad)
{
    Tcl_Obj **padv;
    int i, padc, pixels[4];

    if (TCL_OK != Tcl_ListObjGetElements(interp, objPtr, &padc, &padv)) {
	goto error;
    }

    if (padc > 4) {
	if (interp) {
	    Tcl_ResetResult(interp);
	    Tcl_AppendResult(interp, "Wrong #elements in padding spec", NULL);
	}
	goto error;
    }

    for (i=0; i < padc; ++i) {
	if (Tk_GetPixelsFromObj(interp, tkwin, padv[i], &pixels[i]) != TCL_OK) {
	    goto error;
	}
    }

    /*
     * Fill in defaults:
     */
    switch (padc)
    {
	case 1:	pixels[1] = pixels[0]; /*FALLTHRU*/
	case 2:	pixels[2] = pixels[0]; /*FALLTHRU*/
	case 3:	pixels[3] = pixels[1]; /*FALLTHRU*/
    }

    pad->left	= (short)pixels[0];
    pad->top	= (short)pixels[1];
    pad->right	= (short)pixels[2];
    pad->bottom	= (short)pixels[3];

    return TCL_OK;

error:
    pad->left = pad->top = pad->right = pad->bottom = 0;
    return TCL_ERROR;
}


/*
 * TTK_MakePadding --
 *	Return an initialized TTK_Padding structure.
 */
TTK_Padding TTK_MakePadding(short left, short top, short right, short bottom)
{
    TTK_Padding pad;
    pad.left = left;
    pad.top = top;
    pad.right = right;
    pad.bottom = bottom;
    return pad;
}

/*
 * TTK_UniformPadding --
 * 	Returns a uniform TTK_Padding structure, with the same
 * 	border width on all sides.
 */
TTK_Padding TTK_UniformPadding(short borderWidth)
{
    TTK_Padding pad;
    pad.left = pad.top = pad.right = pad.bottom = borderWidth;
    return pad;
}

/*
 * TTK_AddPadding --
 *	Combine padding 
 */
TTK_Padding TTK_AddPadding(TTK_Padding p1, TTK_Padding p2)
{
    p1.left += p2.left;
    p1.top += p2.top;
    p1.right += p2.right;
    p1.bottom += p2.bottom;
    return p1;
}

/* TTK_RelievePadding --
 * 	Add an extra n pixels of padding according to specified relief.
 * 	This may be used in element geometry procedures to simulate 
 * 	a "pressed-in" look for pushbuttons.
 */
TTK_Padding TTK_RelievePadding(TTK_Padding padding, int relief, int n)
{
    switch (relief)
    {
	case TK_RELIEF_RAISED:
	    padding.right += n;
	    padding.bottom += n;
	    break;
	case TK_RELIEF_SUNKEN:	/* shift */
	    padding.left += n;
	    padding.top += n;
	    break;
	default:
	{
	    int h1 = n/2, h2 = h1 + n % 2;
	    padding.left += h1;
	    padding.top += h1;
	    padding.right += h2;
	    padding.bottom += h2;
	    break;
	}
    }
    return padding;
}

/*
 * TTK_GetStickyFromObj --
 * 	Returns a stickiness specification from the specified Tcl_Obj*,
 * 	consisting of any combination of n, s, e, and w.
 *
 * Returns: TCL_OK, or TCL_ERROR.  If interp is non-NULL
 */
int TTK_GetStickyFromObj(Tcl_Interp *interp, Tcl_Obj *objPtr, unsigned *result)
{
    const char *string = Tcl_GetString(objPtr);
    unsigned sticky = 0;
    char c;

    while ((c = *string++) != '\0') {
	switch (c) {
	    case 'w': case 'W': sticky |= TTK_STICK_W; break;
	    case 'e': case 'E': sticky |= TTK_STICK_E; break;
	    case 'n': case 'N': sticky |= TTK_STICK_N; break;
	    case 's': case 'S': sticky |= TTK_STICK_S; break;
	    default:
	    	if (interp) {
		    Tcl_ResetResult(interp);
		    Tcl_AppendResult(interp, 
			"Bad -sticky specification ",
			Tcl_GetString(objPtr),
			NULL);
		}
		return TCL_ERROR;
	}
    }

    *result = sticky; 
    return TCL_OK;
}


/*
 * +++ TTK_Layout utilities:
 */

struct TTK_LayoutNode_
{
    unsigned		pack;
    TTK_Element 	element;
    const char		*name;
    unsigned int 	state;
    short 		width;		/* requested width */
    short		height;		/* requested height */
    TTK_Box 		parcel;		/* allocated parcel */
    TTK_LayoutNode	*next, *child;
};

static TTK_LayoutNode *NewLayoutNode(
    unsigned pack, TTK_Element element, const char *name)
{
    TTK_LayoutNode *node = (TTK_LayoutNode*)ckalloc(sizeof(TTK_LayoutNode));

    node->pack = pack;
    node->element = element;
    node->state = 0u;
    node->name = name;
    node->next = node->child = 0;
    /* parcel uninitialized */

    return node;
}

void TTK_FreeLayout(TTK_LayoutNode *node)
{
    while (node) {
	TTK_LayoutNode *next = node->next;
	TTK_FreeLayout(node->child);
	ckfree((char*)node);
	node = next;
    }
}

const char *TTK_LayoutNodeName(TTK_LayoutNode *node)
{
    return node->name;
}

TTK_Box TTK_LayoutNodeParcel(TTK_LayoutNode *node)
{
    return node->parcel;
}

void TTK_LayoutNodeSetParcel(TTK_LayoutNode *node, TTK_Box parcel)
{
    node->parcel = parcel;
    /* TODO: relayout children */
}

TTK_Element TTK_LayoutNodeElement(TTK_LayoutNode *node)
{
    return node->element;
}

/*
 * TTK_LayoutNodeInternalParcel --
 * 	Returns the inner area of a specified layout node,
 * 	based on current parcel and element's internal padding.
 */
TTK_Box TTK_LayoutNodeInternalParcel(
    TTK_LayoutNode *node, void *recordPtr, Tk_Window tkwin)
{
    int width, height;
    TTK_Padding padding;
    TTK_ElementGeometry(node->element,
	    recordPtr, tkwin, 0, &width, &height, &padding);
    return TTK_PadBox(node->parcel, padding);
}

void TTK_ChangeElementState(TTK_LayoutNode *node,unsigned set,unsigned clr)
{
    node->state = (node->state | set) & ~clr;
}

/* TTK_BuildLayout --
 *	Build a layout tree from a layout specification.
 */

TTK_Layout 
TTK_BuildLayout(
    TTK_Theme theme,
    TTK_LayoutSpec spec,
    TTK_Style style,
    Tk_OptionTable optionTable)
{
    TTK_LayoutNode *first = 0, *last = 0, *node;

    for ( ; !(spec->opcode & TTK_LAYOUT_END) ; ++spec)
    {
	if (spec->elementName) 
	{
	    TTK_Element element = 
		TTK_GetElement(theme, style, spec->elementName, optionTable);
	    if (!element) {
		Tcl_Panic("Can't find element %s", spec->elementName);
	    	continue;	/* @@@ error? OOPS: this will crash. */
	    }

	    node = NewLayoutNode(spec->opcode, element, spec->elementName);

	    if (last) {
		last->next = node;
		last = node;
	    } else {
		last = first = node;
	    }
	} 

	if (spec->opcode & TTK_CHILDREN) {
	    last->child = TTK_BuildLayout(theme, spec+1, style, optionTable);
	    while (!(spec->opcode & TTK_LAYOUT_END)) {
		++spec;
	    }
	    break;
	}
    } /* for */

    return first;
}

/*
 * TTK_LayoutReqSize --
 * 	Compute minimum requested width & height of a layout tree
 */
void TTK_LayoutReqSize(
    TTK_Layout layout,
    void *recordPtr,
    Tk_Window tkwin,
    unsigned state, 
    int *widthPtr,
    int *heightPtr)
{
    int width = 0, height = 0;

    for ( ; layout != 0; layout = layout->next)
    {
	int elementWidth, elementHeight;
	TTK_Padding elementPadding;

	TTK_ElementGeometry(
		layout->element, recordPtr, tkwin, state|layout->state,
		&elementWidth, &elementHeight, &elementPadding);

	if (layout->child) {
	    int subWidth = 0, subHeight = 0;

	    TTK_LayoutReqSize(layout->child, recordPtr, tkwin, state,
		    &subWidth, &subHeight);

	    subWidth += TTK_PaddingWidth(elementPadding);
	    subHeight += TTK_PaddingHeight(elementPadding);
	    if (elementWidth < subWidth) elementWidth = subWidth;
	    if (elementHeight < subHeight) elementHeight = subHeight;
	}

	if (layout->pack & (TTK_PACK_TOP|TTK_PACK_BOTTOM)) {
	    height += elementHeight;
	} else {
	    height = MAX(height, elementHeight);
	}

	if (layout->pack & (TTK_PACK_LEFT|TTK_PACK_RIGHT)) {
	    width += elementWidth;
	} else {
	    width = MAX(width, elementWidth);
	}
    }
    *widthPtr = width;
    *heightPtr = height;
}


/*
 * TTK_LayoutGeometry --
 * 	Compute parcel for each node in a layout tree
 * 	according to packing specification and overall size.
 */
void TTK_LayoutGeometry(TTK_Layout layout,
	void *recordPtr, Tk_Window tkwin, unsigned int state,
	int x, int y, int width, int height)
{
    TTK_Box cavity = TTK_MakeBox(x, y, width, height);

    for ( ; layout; layout = layout->next)
    {
	int elementWidth, elementHeight;
	TTK_Padding padding;
	TTK_Box parcel;

	/* @@@ factor this, or cache it: */
	TTK_ElementGeometry(layout->element, recordPtr, tkwin, state,
		    &elementWidth, &elementHeight, &padding);
	if (layout->child) {
	    int subWidth = 0, subHeight = 0;
	    TTK_LayoutReqSize(layout->child, recordPtr, tkwin, state,
		    &subWidth, &subHeight);
	    elementWidth = subWidth + padding.left + padding.right;
	    elementHeight = subHeight + padding.top + padding.bottom;
	}

	/*
	 * Compute parcel:
	 */
	if (layout->pack & TTK_EXPAND) {
	    parcel = cavity;
	} else if (layout->pack & TTK_PACK_TOP) {
	    parcel = packTop(&cavity, elementHeight);
	} else if (layout->pack & TTK_PACK_LEFT) {
	    parcel = packLeft(&cavity, elementWidth);
	} else if (layout->pack & TTK_PACK_BOTTOM) {
	    parcel = packBottom(&cavity, elementHeight);
	} else if (layout->pack & TTK_PACK_RIGHT) {
	    parcel = packRight(&cavity, elementWidth);
	} else {
	    parcel = cavity;
	}

	/*
	 * Place element within parcel:
	 */
	layout->parcel = TTK_StickBox(
		parcel, elementWidth, elementHeight, layout->pack); 

	/*
	 * Layout child nodes:
	 */
	if (layout->child) {
	    TTK_Box childBox = TTK_PadBox(layout->parcel, padding);
	    TTK_LayoutGeometry(layout->child,recordPtr,tkwin, state,
		childBox.x, childBox.y, childBox.width, childBox.height);
	}
    }
}

/*
 * TTK_LayoutDraw --
 * 	Draw a layout tree.
 */
void TTK_LayoutDraw(
    TTK_Layout layout, void *recordPtr, Tk_Window tkwin,
    Drawable d, unsigned int state)
{
    for (; layout ; layout = layout->next) 
    {
	int border = layout->pack & TTK_BORDER;

	if (layout->child && border)
	    TTK_LayoutDraw(layout->child, recordPtr, tkwin, d, state);

	TTK_DrawElement(layout->element, recordPtr, tkwin, d,
		layout->parcel, state | layout->state);

	if (layout->child && !border)
	    TTK_LayoutDraw(layout->child, recordPtr, tkwin, d, state);
    }
}

/*
 * TTK_LayoutIdentify --
 * 	Find the layout node at the specified x,y coordinate.
 */
TTK_LayoutNode *TTK_LayoutIdentify(TTK_Layout layout, int x, int y)
{
    TTK_LayoutNode *closest = NULL;

    for (; layout ; layout = layout->next) {
	if (TTK_BoxContains(layout->parcel, x, y)) {
	    TTK_LayoutNode *childNode = 
		layout->child ? TTK_LayoutIdentify(layout->child, x,y) : 0;
	    if (childNode) {
		closest = childNode;
	    } else {
		closest = layout;
	    }
	}
    }
    return closest;
}

/*
 * tail --
 * 	Return the last component of an element name, e.g.,
 * 	"Scrollbar.thumb" => "thumb"
 */
static const char *tail(const char *elementName)
{
    const char *dot;
    while ((dot=strchr(elementName,'.')) != NULL)
	elementName = dot + 1;
    return elementName;
}

/*
 * TTK_LayoutFindNode --
 * 	Look up a layout node by name.
 */
TTK_LayoutNode *TTK_LayoutFindNode(TTK_Layout layout, const char *nodeName)
{
    for (; layout ; layout = layout->next) {
	if (!strcmp(tail(TTK_LayoutNodeName(layout)), nodeName))
	    return layout;

	if (layout->child) {
	    TTK_LayoutNode *childNode = 
		TTK_LayoutFindNode(layout->child, nodeName);
	    if (childNode) 
		return childNode;
	}
    }
    return 0;
}

/*
 *---------------------------------------------------------------------------
 *
 * TTK_ParseLayoutSpec --
 *	Convert a Tcl list into a layout spec structure.
 *
 * Syntax:
 * 	layoutSpec ::= { elementName ?-option value ...? }+
 *
 */

typedef struct LayoutOp_ {
    char *name;
    unsigned flags;
    struct LayoutOp_ *next, *children;
} LayoutOp;

static LayoutOp *NewLayoutOp(void)
{
    LayoutOp *op = (LayoutOp*)ckalloc(sizeof(*op));
    op->name = 0; 
    op->flags = 0;
    op->next = op->children = 0;
    return op;
}

static LayoutOp *TTK_ParseLayout(Tcl_Interp *interp, Tcl_Obj *objPtr)
{
    enum {  OP_SIDE, OP_STICKY, OP_EXPAND, OP_BORDER, OP_CHILDREN };
    static const char *optStrings[] = {
	    "-side", "-sticky", "-expand", "-border", "-children", NULL };

    /* NB: This must match bit definitions TTK_PACK_LEFT etc. */
    static const char *sideStrings[] = 
	{ "left", "right", "top", "bottom", NULL };

    int i = 0, objc;
    Tcl_Obj **objv;
    LayoutOp *head, *tail;

    if (Tcl_ListObjGetElements(interp, objPtr, &objc, &objv) != TCL_OK)
	return 0;

    head = tail = NewLayoutOp();

    while (i < objc) {
	char *elementName = Tcl_GetString(objv[i]);
	unsigned flags = 0x0, sticky = TTK_STICK_ALL;
	Tcl_Obj *childSpec = 0;

	/*
	 * Parse options:
	 */
	++i;
	while (i < objc) {
	    const char *optName = Tcl_GetString(objv[i]);
	    int option, value;

	    if (optName[0] != '-') 
		break;

	    if (Tcl_GetIndexFromObj(
		    interp, objv[i], optStrings, "option", 0, &option) 
		!= TCL_OK)
	    {
		goto error;
	    }

	    if (++i >= objc) {
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp, 
			"Missing value for option ",Tcl_GetString(objv[i-1]),
			NULL);
		goto error;
	    }

	    switch (option) {
		case OP_SIDE:
		    if (Tcl_GetIndexFromObj(interp, objv[i], sideStrings, 
				"side", 0, &value) != TCL_OK)
		    {
			goto error;
		    }
		    flags |= (TTK_PACK_LEFT << value);

		    break;
		case OP_STICKY:
		    if (TTK_GetStickyFromObj(interp,objv[i],&sticky) != TCL_OK)
			goto error;
		    break;
		case OP_EXPAND: 
		    if (Tcl_GetBooleanFromObj(interp,objv[i],&value) != TCL_OK)
			goto error;
		    if (value) 
			flags |= TTK_EXPAND;
		    break;
		case OP_BORDER: 
		    if (Tcl_GetBooleanFromObj(interp,objv[i],&value) != TCL_OK)
			goto error;
		    if (value) 
			flags |= TTK_BORDER;
		    break;
		case OP_CHILDREN:
		    childSpec = objv[i];
		    break;
	    }
	    ++i;
	}

	/*
	 * Build new node:
	 */
	tail->name = strdup(elementName);	/* @@@ don't use strdup() */
	tail->flags = flags | sticky;
	if (childSpec) {
	    tail->children = TTK_ParseLayout(interp, childSpec);
	    if (!tail->children)
		goto error;
	}

	if (i < objc) {
	    tail->next = NewLayoutOp();
	    tail = tail->next;
	}
    }

    return head;

error:
    /* @@@ free layout tree */
    return 0;
}

/*
 * ParseLayoutSpec --
 * 	Parse a layout tree, reassemble into a layout spec.
 * 	@@@ This is backwards: TTK_RegisterLayout() ought to 
 * 	@@@ turn static layout specs into a tree of LayoutOps instead.
 * 	@@@ We'll do that later; in the meantime, do a half-assed job here.
 */

static int CountLayoutNodes(LayoutOp *op)
{
    int n = 0;
    while (op) {
	++n;
	if (op->children) 
	    n += CountLayoutNodes(op->children);
	op = op->next;
    }
    return n + 1;
}

static int FillSpec(LayoutOp *op, TTKLayoutInstruction *spec, int i)
{
    while (op) {
	spec[i].elementName = op->name;
	spec[i].opcode = op->flags;
	if (op->children) {
	    spec[i].opcode |= TTK_CHILDREN;
	    i = FillSpec(op->children, spec, i + 1);
	}
	++i; op = op->next;
    }
    spec[i].elementName = 0;
    spec[i].opcode = TTK_LAYOUT_END;
    return i;
}

TTK_LayoutSpec TTK_ParseLayoutSpec(Tcl_Interp *interp, Tcl_Obj *objPtr)
{
    LayoutOp *op = TTK_ParseLayout(interp, objPtr);
    TTKLayoutInstruction *specPtr;
    int nnodes = CountLayoutNodes(op);

    if (!op) return 0;
    specPtr = (TTKLayoutInstruction *)
	ckalloc(sizeof(TTKLayoutInstruction) * nnodes);
    FillSpec(op,specPtr, 0);
    return specPtr;
}

/*
 * +++ Miscellaneous stuff:
 */

/*
 * adjustAnchor --
 * 	Adjust the x and y coordinates to anchor the inner box
 * 	inside the outer box according to the specified anchor.
 */
void adjustAnchor(
    Tk_Anchor anchor,
    int *xPtr, int *yPtr, int width, int height,	/* outer box */
    int innerWidth, int innerHeight)
{
    int dx = width - innerWidth;
    int dy = height - innerHeight;

    /*
     * X coordinate adjustment:
     */
    switch (anchor)
    {
	case TK_ANCHOR_NW: case TK_ANCHOR_W: case TK_ANCHOR_SW:
	    break; /* no change */
	case TK_ANCHOR_N: case TK_ANCHOR_CENTER: case TK_ANCHOR_S:
	    *xPtr += dx / 2;
	    break;
	case TK_ANCHOR_NE: case TK_ANCHOR_E: case TK_ANCHOR_SE:
	    *xPtr += dx;
	    break;
    }

    /*
     * Y coordinate adjustment:
     */
    switch (anchor)
    {
	case TK_ANCHOR_NW: case TK_ANCHOR_N: case TK_ANCHOR_NE:
	    break;	/* no change */
	case TK_ANCHOR_W: case TK_ANCHOR_CENTER: case TK_ANCHOR_E:
	    *yPtr += dy / 2;
	    break;
	case TK_ANCHOR_SW: case TK_ANCHOR_S: case TK_ANCHOR_SE:
	    *yPtr += dy;
	    break;
    }
}

/*
 * +++ Helper routines for configuring GCs:
 */

XColor *getForegroundFromObj(
    Tk_Window tkwin, 
    Tcl_Obj *objPtr, 
    unsigned long *maskPtr, XGCValues *valuesPtr)
{
    XColor *color = Tk_GetColorFromObj(tkwin, objPtr);
    if (color) {
	valuesPtr->foreground = color->pixel;
	*maskPtr |= GCForeground;
	return color;
    }
    return 0;
}

/*
 * NOTE: This routine uses Tk_GetFontFromObj(); the caller
 * must arrange to call Tk_AllocFontFromObj() / Tk_FreeFont(). 
 *
 * It's safe to use this routine if objPtr points to
 * a TK_OPTION_FONT_RESOURCE, in which case the widget
 * resource management code takes care of it.
 */
Tk_Font getFontFromObj(
    Tk_Window tkwin, 
    Tcl_Obj *objPtr, 
    unsigned long *maskPtr, XGCValues *valuesPtr)
{
    Tk_Font tkfont = Tk_GetFontFromObj(tkwin, objPtr);
    if (tkfont) {
	valuesPtr->font = Tk_FontId(tkfont);
	*maskPtr |= GCFont;
	return tkfont;
    }
    return 0;
}

/*EOF*/
