/* scrollbar.c,v 1.32 2004/07/05 17:54:19 jenglish Exp
 *
 * Copyright (c) 2003, Joe English
 *
 * Scrollbar widget implementation.
 *
 */

#include <string.h>
#include <tk.h>

#include "tkTheme.h"
#include "widget.h"
#include "compat.h"

/*
 * Scrollbar widget record:
 */
typedef struct
{
    Tcl_Obj	*commandObj;

    int 	orient;
    Tcl_Obj	*orientObj;

    double	first;			/* top fraction */
    double	last;			/* bottom fraction */

    TTK_Box	troughBox;		/* trough parcel */ 
    int 	minSize;		/* minimum size of thumb */
} ScrollbarPart;

typedef struct
{
    WidgetCore core;
    ScrollbarPart scrollbar;
    SCROLLBAR_COMPAT_DECLS
} Scrollbar;

static Tk_OptionSpec ScrollbarOptionSpecs[] =
{
    {TK_OPTION_STRING, "-command", "command", "Command", "",
	Tk_Offset(Scrollbar,scrollbar.commandObj), -1, 0,0,0},

    {TK_OPTION_STRING_TABLE, "-orient", "orient", "Orient", "vertical",
	Tk_Offset(Scrollbar,scrollbar.orientObj),
	Tk_Offset(Scrollbar,scrollbar.orient),
	0,(ClientData)TTKOrientStrings,STYLE_CHANGED },

    WIDGET_INHERIT_OPTIONS(CoreOptionSpecs)
};

SCROLLBAR_COMPAT_OPTIONS

static int 
ScrollbarInitialize(Tcl_Interp *interp, void *recordPtr)
{
    Scrollbar *sb = recordPtr;
    sb->scrollbar.first = 0.0;
    sb->scrollbar.last = 1.0;

    TrackElementState(&sb->core);

    return TCL_OK;
}

static TTK_Layout ScrollbarGetLayout(
    Tcl_Interp *interp, TTK_Theme theme, void *recordPtr)
{
    Scrollbar *sb = recordPtr;
    return TTK_CreateLayout( interp, theme,
	    sb->scrollbar.orient == TTK_ORIENT_HORIZONTAL 
		?  "Horizontal.TScrollbar" : "Vertical.TScrollbar",
	    sb->core.optionTable);
}

/*
 * ScrollbarLayout --
 * 	Computes the position of the scrollbar thumb, along with
 * 	the rest of te
 * 	
 * Side effects:
 * 	Sets sb->troughBox and sb->minSize.
 */
static void ScrollbarDoLayout(void *recordPtr)
{
    Scrollbar *sb = recordPtr;
    WidgetCore *corePtr = &sb->core;
    Tk_Window tkwin = corePtr->tkwin;
    TTK_LayoutNode *thumb;
    TTK_Box thumbBox;
    int thumbWidth, thumbHeight;
    TTK_Padding thumbPadding;
    double first, last, size;
    int minSize;

    /*
     * Use generic layout manager to compute initial layout:
     */
    TTK_LayoutGeometry(corePtr->layout, (ClientData)sb, tkwin, corePtr->state,
	    0,0, Tk_Width(tkwin), Tk_Height(tkwin));

    /*
     * Locate thumb element:
     */
    thumb = TTK_LayoutFindNode(corePtr->layout, "thumb");
    if (!thumb)	/* Something has gone wrong -- bail */
	return;

    sb->scrollbar.troughBox = thumbBox = TTK_LayoutNodeParcel(thumb);

    /*
     * Query the thumb element for its minimum size:
     */
    TTK_ElementGeometry(TTK_LayoutNodeElement(thumb),
	    (ClientData)sb, tkwin, corePtr->state, 
	    &thumbWidth, &thumbHeight, &thumbPadding);

    /*
     * Adjust thumb element parcel:
     */
    first = sb->scrollbar.first;
    last  = sb->scrollbar.last;

    if (sb->scrollbar.orient == TTK_ORIENT_VERTICAL) {
	minSize = thumbHeight;
	size = thumbBox.height - minSize;
	thumbBox.y += (int)(size * first);
	thumbBox.height = (int)(size * last) + minSize - (int)(size * first);
    } else {
	minSize = thumbWidth;
	size = thumbBox.width - minSize;
	thumbBox.x += (int)(size * first);
	thumbBox.width = (int)(size * last) + minSize - (int)(size * first);
    }
    sb->scrollbar.minSize = minSize;
    TTK_LayoutNodeSetParcel(thumb, thumbBox);
}

/*
 * Implementation of scrollbar 'set' widget method:
 * 	$sb set $first $last
 */
static int
ScrollbarSetCommand(
    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], void *recordPtr)
{
    Scrollbar *scrollbar = recordPtr;
    Tcl_Obj *firstObj, *lastObj;
    double first, last;

    if (objc != 4) {
	Tcl_WrongNumArgs(interp, 2, objv, "first last");
	return TCL_ERROR;
    }

    firstObj = objv[2];
    lastObj = objv[3];
    if (Tcl_GetDoubleFromObj(interp, firstObj, &first) != TCL_OK
	|| Tcl_GetDoubleFromObj(interp, lastObj, &last) != TCL_OK)
	return TCL_ERROR;

    /* Range-checks:
     */
    if (first < 0.0)
	first = 0.0;
    else if (first > 1.0)
	first = 1.0;

    if (last < first)
	last = first;
    else if (last > 1.0)
	last = 1.0;

    /* ASSERT: 0.0 <= first <= last <= 1.0 */

    scrollbar->scrollbar.first = first;
    scrollbar->scrollbar.last = last;

    WidgetChanged(&scrollbar->core, REDISPLAY_REQUIRED);

    return TCL_OK;
}

/*
 * Scrollbar 'get' method:
 * Returns the last thing passed to -set
 */
static int
ScrollbarGetCommand(
    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], void *recordPtr)
{
    Scrollbar *scrollbar = recordPtr;
    Tcl_Obj *result;

    if (objc != 2) {
	Tcl_WrongNumArgs(interp, 2, objv, "");
	return TCL_ERROR;
    }

    result = Tcl_NewListObj(0,0);
    Tcl_ListObjAppendElement(interp, result, 
	    Tcl_NewDoubleObj(scrollbar->scrollbar.first));
    Tcl_ListObjAppendElement(interp, result, 
	    Tcl_NewDoubleObj(scrollbar->scrollbar.last));
    Tcl_SetObjResult(interp, result);
    return TCL_OK;
}

/*
 * Scrollbar 'delta' method:
 * $sb delta $dx $dy
 * Returns the percentage change corresponding to a mouse movement
 * of $dx, $dy.
 */

static int
ScrollbarDeltaCommand(
    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], void *recordPtr)
{
    Scrollbar *sb = recordPtr;
    double dx, dy;
    double delta = 0.0;

    if (objc != 4) {
	Tcl_WrongNumArgs(interp, 2, objv, "dx dy");
	return TCL_ERROR;
    }

    if (Tcl_GetDoubleFromObj(interp, objv[2], &dx) != TCL_OK
	|| Tcl_GetDoubleFromObj(interp, objv[3], &dy) != TCL_OK)
    {
	return TCL_ERROR;
    }

    delta = 0.0;
    if (sb->scrollbar.orient == TTK_ORIENT_VERTICAL) {
	int size = sb->scrollbar.troughBox.height - sb->scrollbar.minSize;
	if (size > 0)
	    delta = (double)dy / (double)size;
    } else {
	int size = sb->scrollbar.troughBox.width - sb->scrollbar.minSize;
	if (size > 0)
	    delta = (double)dx / (double)size;
    }

    Tcl_SetObjResult(interp, Tcl_NewDoubleObj(delta));
    return TCL_OK;
}

/*
 * Scrollbar 'fraction' method:
 * $sb fraction $x $y
 * Returns a real number between 0 and 1 indicating  where  the
 * point given by x and y lies in the trough area of the scrollbar. 
 */
static int
ScrollbarFractionCommand(
    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], void *recordPtr)
{
    Scrollbar *sb = recordPtr;
    TTK_Box b = sb->scrollbar.troughBox;
    int minSize = sb->scrollbar.minSize;
    double x, y;
    double fraction = 0.0;

    if (objc != 4) {
	Tcl_WrongNumArgs(interp, 2, objv, "x y");
	return TCL_ERROR;
    }

    if (Tcl_GetDoubleFromObj(interp, objv[2], &x) != TCL_OK
	|| Tcl_GetDoubleFromObj(interp, objv[3], &y) != TCL_OK)
    {
	return TCL_ERROR;
    }

    fraction = 0.0;
    if (sb->scrollbar.orient == TTK_ORIENT_VERTICAL) {
	if (b.height != 0)
	    fraction = (double)(y - b.y) / (double)(b.height - minSize);
    } else {
	if (sb->scrollbar.troughBox.width != 0)
	    fraction = (double)(x - b.x) / (double)(b.width - minSize);
    }

    Tcl_SetObjResult(interp, Tcl_NewDoubleObj(fraction));
    return TCL_OK;
}

static WidgetCommandSpec ScrollbarCommands[] =
{
    { "configure",	WidgetConfigureCommand },
    { "cget",		WidgetCgetCommand },
    { "delta",    	ScrollbarDeltaCommand },
    { "fraction",    	ScrollbarFractionCommand },
    { "get",    	ScrollbarGetCommand },
    { "identify",	WidgetIdentifyCommand },
    { "instate",	WidgetInstateCommand },
    { "set",  		ScrollbarSetCommand },
    { "state",  	WidgetStateCommand },
    { 0,0 }
};

/*
 * Widget specification:
 */
WidgetSpec ScrollbarWidgetSpec =
{
    "TScrollbar",		/* className */
    sizeof(Scrollbar),		/* recordSize */
    ScrollbarCompatOptionSpecs,	/* optionSpecs */
    ScrollbarCommands,		/* subcommands */
    ScrollbarInitialize,	/* initializeProc */
    NullCleanup,		/* cleanupProc */
    CoreConfigure,		/* configureProc */
    NullPostConfigure,		/* postConfigureProc */
    ScrollbarGetLayout,		/* getLayoutProc */
    WidgetSize, 		/* sizeProc */
    ScrollbarDoLayout,		/* layoutProc */
    WidgetDisplay,		/* displayProc */

    WIDGET_SPEC_END		/* sentinel */
};

