/*
 * aquaTheme.c,v 1.7 2004/07/05 18:48:53 jenglish Exp
 *
 * Tk theme engine for Mac OSX, using the Appearance Manager API.
 *
 * Copyright (c) 2004 Joe English
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * See also:
 *
 * <URL: http://developer.apple.com/documentation/Carbon/Reference/
 * 	Appearance_Manager/appearance_manager/APIIndex.html >
 *
 * Notes:
 * 	"Active" means different things in Mac and Tk terminology --
 *	On Aqua, widgets are "Active" if they belong to the foreground window,
 *	"Inactive" if they are in a background window.
 *	Tk/tile uses the term "active" to mean that the mouse cursor
 *	is over a widget; aka "hover", "prelight", or "hot-tracked".
 *	(Aqua doesn't use this kind of feedback).
 *
 * 	The QuickDraw/Carbon coordinate system is relative to the
 *	top-level window, *not* to the Tk_Window.  However,
 *	since we're drawing into an off-screen port (Tk "Pixmap),
 *	we don't need to account for this.
 */

/* @@@ Workaround for busted system headers on Alicia's laptop: */
#define AVAILABLE_MAC_OS_X_VERSION_10_3_AND_LATER

#include <Carbon/Carbon.h>
#include <tk.h>
#include <tkMacOSX.h>
#include "tkTheme.h"

/*----------------------------------------------------------------------
 * +++ Utilities.
 */

Rect BoxToRect(TTK_Box b)
{
    Rect rect;
    rect.top	= b.y;
    rect.left	= b.x;
    rect.bottom	= b.y + b.height;
    rect.right	= b.x + b.width;
    return rect;
}

/* Table mapping Tk states to Appearance manager ThemeStates
 */

static TTK_StateTable ThemeStateTable[] = {
    {kThemeStateUnavailable, TTK_STATE_DISABLED, 0},
    {kThemeStatePressed, TTK_STATE_PRESSED, 0},
    {kThemeStateInactive, TTK_STATE_BACKGROUND, 0},
    {kThemeStateActive, 0, 0}
/* Others: Not sure what these are supposed to mean.
   Up/Down have something to do with "little arrow" increment controls...
   Dunno what a "Rollover" is.
    {kThemeStateRollover, 0, 0},
    {kThemeStateUnavailableInactive, 0, 0}
    {kThemeStatePressedUp, 0, 0},
    {kThemeStatePressedDown, 0, 0}
*/
};

/*----------------------------------------------------------------------
 * +++ Button element:  Used for elements drawn with DrawThemeButton.
 */

/* Extra margins to account for drop shadow.
 */
static TTK_Padding ButtonMargins = {2,2,2,2};

#define NoThemeMetric 0xFFFFFFFF

typedef struct {
    ThemeButtonKind	kind;
    ThemeMetric 	heightMetric;
} ThemeButtonParms;

enum {
    KTHEMEPushButton,
    KTHEMECheckBox,
    KTHEMERadioButton,
    KTHEMEBevelButton,
    KTHEMEPopupButton };

static ThemeButtonParms ButtonParms[] = {
    { kThemePushButton, NoThemeMetric },
    { kThemeCheckBox, kThemeMetricCheckBoxHeight },
    { kThemeRadioButton, kThemeMetricRadioButtonHeight },
    { kThemeBevelButton, NoThemeMetric },
    { kThemePopupButton, NoThemeMetric },
};

static TTK_StateTable ButtonValueTable[] = {
    { kThemeButtonMixed, TTK_STATE_INDETERMINATE, 0 },
    { kThemeButtonOn, TTK_STATE_SELECTED, 0 },
    { kThemeButtonOff, 0, 0 }
/* Others: kThemeDisclosureRight, kThemeDisclosureDown, kThemeDisclosureLeft */
};

static TTK_StateTable ButtonAdornmentTable[] = {
    { kThemeAdornmentDefault, TTK_STATE_DEFAULT, 0 },
    { kThemeAdornmentFocus, TTK_STATE_FOCUS, 0 },
    { kThemeAdornmentNone, 0, 0 }
};

/*
 * computeButtonDrawInfo --
 *	Fill in an appearance manager ThemeButtonDrawInfo record.
 */
static ThemeButtonDrawInfo computeButtonDrawInfo(
	ThemeButtonParms *parms, unsigned state)
{
    ThemeButtonDrawInfo info;
    info.state = TTK_StateTableLookup(ThemeStateTable, state);
    info.value = TTK_StateTableLookup(ButtonValueTable, state);
    info.adornment = TTK_StateTableLookup(ButtonAdornmentTable, state);
    return info;
}

static void ButtonElementGeometry(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, int *widthPtr, int *heightPtr, TTK_Padding *paddingPtr)
{
    ThemeButtonParms *parms = clientData;
    ThemeButtonDrawInfo drawInfo = computeButtonDrawInfo(parms, 0);
    Rect scratchRect, contentsRect;
    const int scratchSize = 100;

    if (parms->heightMetric != NoThemeMetric) {
	SInt32 gratuitouslyOverspecifiedType;
	GetThemeMetric(parms->heightMetric, &gratuitouslyOverspecifiedType);
	*heightPtr = gratuitouslyOverspecifiedType;
    }

    /* To compute internal padding, query the appearance manager
     * for the content bounds of a dummy rectangle, then use
     * the difference as the padding.
     */
    scratchRect.top = scratchRect.left = 0;
    scratchRect.bottom = scratchRect.right = scratchSize;

    GetThemeButtonContentBounds(
	&scratchRect, parms->kind, &drawInfo, &contentsRect);

    paddingPtr->left = contentsRect.left;
    paddingPtr->top = contentsRect.top;
    paddingPtr->bottom = scratchSize - contentsRect.bottom;
    paddingPtr->right = scratchSize - contentsRect.right;

    /* Now add a little extra padding to account for drop shadows.
     * @@@ SHOULD: call GetThemeButtonBackgroundBounds() instead.
     */

    *paddingPtr = TTK_AddPadding(*paddingPtr, ButtonMargins);
}

static void
ButtonElementDraw(void *clientData, void *elementRecord,
    Tk_Window tkwin, Drawable d, TTK_Box b, unsigned int state)
{
    Rect bounds = BoxToRect(TTK_PadBox(b, ButtonMargins));
    ThemeButtonParms *parms = clientData;
    ThemeButtonDrawInfo info = computeButtonDrawInfo(parms, state);

    SetGWorld(TkMacOSXGetDrawablePort(d), 0);
    DrawThemeButton(&bounds, parms->kind, &info,
	NULL/*prevInfo*/,NULL/*eraseProc*/,NULL/*labelProc*/,NULL/*userData*/);
}

static TTK_ElementSpec ButtonElementSpec =
{
    TK_STYLE_VERSION_2,
    sizeof(NullElement),
    NullElementOptions,
    ButtonElementGeometry,
    ButtonElementDraw
};

/*----------------------------------------------------------------------
 * +++ Notebook elements.
 */

static TTK_StateTable TabStyleTable[] = {
    { kThemeTabFrontInactive, TTK_STATE_SELECTED|TTK_STATE_BACKGROUND, 0 },
    { kThemeTabNonFrontInactive, TTK_STATE_BACKGROUND, 0 },
    { kThemeTabFrontUnavailable,  TTK_STATE_DISABLED|TTK_STATE_SELECTED, 0 },
    { kThemeTabNonFrontUnavailable, TTK_STATE_DISABLED, 0 },
    { kThemeTabFront, TTK_STATE_SELECTED, 0 },
    { kThemeTabNonFrontPressed, TTK_STATE_PRESSED, 0 },
    { kThemeTabNonFront, 0,0 }
};

/* Quoth DrawThemeTab() reference manual:
 * "Small tabs have a height of 16 pixels large tabs have a height of
 * 21 pixels. (The widths of tabs are variable.) Additionally, the
 * distance that the tab overlaps the pane must be included in the tab
 * rectangle this overlap distance is always 3 pixels, although the
 * 3-pixel overlap is only drawn for the front tab."
 */
static const int TAB_HEIGHT = 21;
static const int TAB_OVERLAP = 3;

static void TabElementGeometry(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, int *widthPtr, int *heightPtr, TTK_Padding *paddingPtr)
{
    *heightPtr = TAB_HEIGHT + TAB_OVERLAP - 1;
}

static void
TabElementDraw(void *clientData, void *elementRecord,
    Tk_Window tkwin, Drawable d, TTK_Box b, unsigned int state)
{
    Rect bounds = BoxToRect(b);
    bounds.bottom += TAB_OVERLAP;
    SetGWorld(TkMacOSXGetDrawablePort(d), 0);
    DrawThemeTab(
    	&bounds, TTK_StateTableLookup(TabStyleTable, state), kThemeTabNorth,
	0/*labelProc*/,0/*userData*/);
}

static TTK_ElementSpec TabElementSpec =
{
    TK_STYLE_VERSION_2,
    sizeof(NullElement),
    NullElementOptions,
    TabElementGeometry,
    TabElementDraw
};

/* Notebook panes:
 */
static void PaneElementGeometry(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, int *widthPtr, int *heightPtr, TTK_Padding *paddingPtr)
{
    /* Padding determined by trial-and-error */
    *paddingPtr = TTK_MakePadding(2,8,2,2);
}

static void
PaneElementDraw(void *clientData, void *elementRecord,
    Tk_Window tkwin, Drawable d, TTK_Box b, unsigned int state)
{
    Rect bounds = BoxToRect(b);
    SetGWorld(TkMacOSXGetDrawablePort(d), 0);
    DrawThemeTabPane(
    	&bounds, TTK_StateTableLookup(ThemeStateTable, state));
}

static TTK_ElementSpec PaneElementSpec =
{
    TK_STYLE_VERSION_2,
    sizeof(NullElement),
    NullElementOptions,
    PaneElementGeometry,
    PaneElementDraw
};

/* Labelframe borders:
 * Use "primary group box ..."
 * Quoth DrawThemePrimaryGroup reference:
 * "The primary group box frame is drawn inside the specified
 * rectangle and is a maximum of 2 pixels thick."
 *
 * "Maximum of 2 pixels thick" is apparently a lie;
 * looks more like 4 to me with shading.
 */
static void GroupElementGeometry(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, int *widthPtr, int *heightPtr, TTK_Padding *paddingPtr)
{
    *paddingPtr = TTK_UniformPadding(4);
}

static void
GroupElementDraw(void *clientData, void *elementRecord,
    Tk_Window tkwin, Drawable d, TTK_Box b, unsigned int state)
{
    Rect bounds = BoxToRect(b);
    SetGWorld(TkMacOSXGetDrawablePort(d), 0);
    DrawThemePrimaryGroup(
    	&bounds, TTK_StateTableLookup(ThemeStateTable, state));
}

static TTK_ElementSpec GroupElementSpec =
{
    TK_STYLE_VERSION_2,
    sizeof(NullElement),
    NullElementOptions,
    GroupElementGeometry,
    GroupElementDraw
};

/*----------------------------------------------------------------------
 * +++ Entry element -- 
 * 	3 pixels padding for focus rectangle
 *	2 pixels padding for EditTextFrame 
 */

typedef struct {
    Tcl_Obj	*backgroundObj;
} EntryElement;

static TTK_ElementOptionSpec EntryElementOptions[] = {
    { "-background", TK_OPTION_BORDER,
	    Tk_Offset(EntryElement,backgroundObj), "white" },
    {0}
};

static void EntryElementGeometry(
    void *clientData, void *elementRecord, Tk_Window tkwin,
    int *widthPtr, int *heightPtr, TTK_Padding *paddingPtr)
{
    *paddingPtr = TTK_UniformPadding(5);
}

static void EntryElementDraw(
    void *clientData, void *elementRecord, Tk_Window tkwin,
    Drawable d, TTK_Box b, unsigned int state)
{
    EntryElement *e = elementRecord;
    Tk_3DBorder backgroundPtr = Tk_Get3DBorderFromObj(tkwin,e->backgroundObj);
    TTK_Box inner = TTK_PadBox(b, TTK_UniformPadding(3));
    Rect bounds = BoxToRect(inner);

    SetGWorld(TkMacOSXGetDrawablePort(d), 0);

    /* Erase w/background color:
    */
    XFillRectangle(Tk_Display(tkwin), d,
	Tk_3DBorderGC(tkwin, backgroundPtr, TK_3D_FLAT_GC),
	inner.x,inner.y, inner.width, inner.height);

    /* Draw border:
    */
    DrawThemeEditTextFrame(
    	&bounds, TTK_StateTableLookup(ThemeStateTable, state));

    /* Draw focus highlight:
    */
    if (state & TTK_STATE_FOCUS)
	DrawThemeFocusRect(&bounds, 1);
}

static TTK_ElementSpec EntryElementSpec =
{
    TK_STYLE_VERSION_2,
    sizeof(EntryElement),
    EntryElementOptions,
    EntryElementGeometry,
    EntryElementDraw
};

/*----------------------------------------------------------------------
 * +++ Pop-up arrow (for comboboxes)
 * NOTE: This isn't right at all, but I can't find the correct
 * function in the Appearance Manager reference.
 */

static void PopupArrowElementGeometry(
    void *clientData, void *elementRecord, Tk_Window tkwin, 
    int *widthPtr, int *heightPtr, TTK_Padding *paddingPtr)
{
    *widthPtr = 12;	/* wild-assed guess */
}

static void PopupArrowElementDraw(
    void *clientData, void *elementRecord, Tk_Window tkwin, 
    Drawable d, TTK_Box b, unsigned int state)
{
    Rect bounds = BoxToRect(b);

    SetGWorld(TkMacOSXGetDrawablePort(d), 0);
    DrawThemePopupArrow(&bounds,
    	kThemeArrowDown, 
	kThemeArrow9pt,		/* ??? */
	TTK_StateTableLookup(ThemeStateTable, state),
	NULL /*eraseProc*/,
	NULL /*eraseData*/);
}

static TTK_ElementSpec PopupArrowElementSpec = 
{
    TK_STYLE_VERSION_2,
    sizeof(NullElement),
    NullElementOptions,
    PopupArrowElementGeometry,
    PopupArrowElementDraw
};

/*----------------------------------------------------------------------
 * +++ DrawThemeTrack-based elements --
 * Progress bars and scales.  (See also: <<NOTE-TRACKS>>)
 */

static TTK_StateTable ThemeTrackEnableTable[] = {
    { kThemeTrackDisabled, TTK_STATE_DISABLED, 0 },
    { kThemeTrackInactive, TTK_STATE_BACKGROUND, 0 },
    { kThemeTrackActive, 0, 0 }
    /* { kThemeTrackNothingToScroll, ?, ? }, */
};

typedef struct {	/* TrackElement client data */
    ThemeTrackKind	kind;
    SInt32		thicknessMetric;
} TrackElementData;

static TrackElementData ProgressData =
    { kThemeProgressBar, kThemeMetricLargeProgressBarThickness };
static TrackElementData ScaleData =
    { kThemeSlider, kThemeMetricHSliderHeight };

typedef struct {
    Tcl_Obj *fromObj;		/* minimum value */
    Tcl_Obj *toObj;		/* maximum value */
    Tcl_Obj *valueObj;		/* current value */
    Tcl_Obj *orientObj; 	/* horizontal / vertical */
} TrackElement;

static TTK_ElementOptionSpec TrackElementOptions[] = {
    { "-from", TK_OPTION_DOUBLE, Tk_Offset(TrackElement,fromObj) },
    { "-to", TK_OPTION_DOUBLE, Tk_Offset(TrackElement,toObj) },
    { "-value", TK_OPTION_DOUBLE, Tk_Offset(TrackElement,valueObj) },
    { "-orient", TK_OPTION_STRING, Tk_Offset(TrackElement,orientObj) },
    {0,0,0}
};

static void
TrackElementGeometry(
    void *clientData, void *elementRecord, Tk_Window tkwin,
    int *widthPtr, int *heightPtr, TTK_Padding *paddingPtr)
{
    TrackElementData *data = clientData;
    SInt32 size = 24;	/* reasonable default ... */
    GetThemeMetric(data->thicknessMetric, &size);
    *widthPtr = *heightPtr = size;
}

static void
TrackElementDraw(
    void *clientData, void *elementRecord, Tk_Window tkwin,
    Drawable d, TTK_Box b, unsigned int state)
{
    TrackElementData *data = clientData;
    TrackElement *elem = elementRecord;
    double from = 0, to = 100, value = 0;
    int orientation = TTK_ORIENT_HORIZONTAL;
    ThemeTrackDrawInfo drawInfo;

    Tcl_GetDoubleFromObj(NULL, elem->fromObj, &from);
    Tcl_GetDoubleFromObj(NULL, elem->toObj, &to);
    Tcl_GetDoubleFromObj(NULL, elem->valueObj, &value);
    TTK_GetOrientationFromObj(NULL, elem->orientObj, &orientation);

    /* @@@ BUG: min, max, and value should account for resolution:
     * @@@ if finer than 1.0, conversion to int breaks.
     */
    drawInfo.kind = data->kind;
    drawInfo.bounds = BoxToRect(b);
    drawInfo.min = (int)from;		/* @@@ */
    drawInfo.max = (int)to;		/* @@@ */
    drawInfo.value = (int)value;	/* @@@ */

    drawInfo.attributes = orientation == TTK_ORIENT_HORIZONTAL
	? kThemeTrackHorizontal : 0;
    drawInfo.attributes |= kThemeTrackShowThumb;
    drawInfo.enableState = TTK_StateTableLookup(ThemeTrackEnableTable, state);

    switch (data->kind) {
    case kThemeProgressBar:
	drawInfo.trackInfo.progress.phase = 0;	/* 1-4: animation phase */
	break;
    case kThemeSlider:
    	drawInfo.trackInfo.slider.pressState = 0; /* @@@ fill this in */
    	drawInfo.trackInfo.slider.thumbDir =  kThemeThumbDownward;
	    /* kThemeThumbUpward, kThemeThumbDownward, kThemeThumbPlain  */
	break;
    }

    SetGWorld(TkMacOSXGetDrawablePort(d), 0);
    DrawThemeTrack(&drawInfo,
    	NULL/*rgnGhost*/,NULL/*eraseProc*/,0/*eraseData*/);
}

static TTK_ElementSpec TrackElementSpec = {
    TK_STYLE_VERSION_2,
    sizeof(TrackElement),
    TrackElementOptions,
    TrackElementGeometry,
    TrackElementDraw
};


/* Slider element -- <<NOTE-TRACKS>>
 * Has geometry only.  The Scale widget adjusts the position of this element,
 * and uses it for hit detection.  In the Aqua theme, the slider is actually
 * drawn as part of the trough element.
 *
 * Also buggy: The geometry here is a Wild-Assed-Guess; I can't
 * figure out how to get the Appearance Manager to tell me the
 * slider size.
 */
static void SliderElementGeometry(
    void *clientData, void *elementRecord, Tk_Window tkwin,
    int *widthPtr, int *heightPtr, TTK_Padding *paddingPtr)
{
	*widthPtr = *heightPtr = 24;
}

static TTK_ElementSpec SliderElementSpec = {
    TK_STYLE_VERSION_2,
    sizeof(NullElement),
    NullElementOptions,
    SliderElementGeometry,
    NullElementDraw
};

/*----------------------------------------------------------------------
 * +++ Background element -- an experiment.
 *
 *	This isn't quite right: In Aqua, the correct background for
 *	a control depends on what kind of container it belongs to,
 *	and the type of the top-level window.
 *
 *	Also: patterned backgrounds should be aligned with the coordinate
 *	system of the top-level window.  Since we're drawing into an
 * 	off-screen graphics port with its own coordinate system,
 *	this leads to alignment glitches.
 *
 *	Available kTheme constants:
 *  	kThemeBackgroundTabPane,
 *  	kThemeBackgroundPlacard,
 *  	kThemeBackgroundWindowHeader,
 *  	kThemeBackgroundListViewWindowHeader,
 *  	kThemeBackgroundSecondaryGroupBox,
 *
 * 	GetThemeBrush() and SetThemeBackground() offer more choices.
 *
 */

static void BackgroundElementDraw(
    void *clientData, void *elementRecord, Tk_Window tkwin,
    Drawable d, TTK_Box b, unsigned int state)
{
    ThemeBackgroundKind kind = kThemeBackgroundWindowHeader;
    Rect bounds;
    SInt32 depth = 32;	/* ??? */
    Boolean inColor = true;

    bounds.top = bounds.left = 0;
    bounds.right = Tk_Width(tkwin);
    bounds.bottom = Tk_Height(tkwin);

    SetGWorld(TkMacOSXGetDrawablePort(d), 0);
    ApplyThemeBackground(kind, &bounds,
    	TTK_StateTableLookup(ThemeStateTable, state),
	depth, inColor);
    EraseRect(&bounds);
}

static TTK_ElementSpec BackgroundElementSpec =
{
    TK_STYLE_VERSION_2,
    sizeof(NullElement),
    NullElementOptions,
    NullElementGeometry,
    BackgroundElementDraw
};

/*----------------------------------------------------------------------
 * +++ Widget layouts.
 */

TTK_BEGIN_LAYOUT(ButtonLayout)
    TTK_NODE("Button.background", TTK_FILL_BOTH)
    TTK_GROUP("Button.button", TTK_FILL_BOTH,
	TTK_GROUP("Button.padding", TTK_FILL_BOTH,
	    TTK_NODE("Button.label", TTK_FILL_BOTH)))
TTK_END_LAYOUT

TTK_BEGIN_LAYOUT(RadiobuttonLayout)
    TTK_NODE("Radiobutton.background", TTK_FILL_BOTH)
    TTK_GROUP("Radiobutton.button", TTK_FILL_BOTH,
	TTK_GROUP("Radiobutton.padding", TTK_FILL_BOTH|TTK_EXPAND,
	    TTK_NODE("Radiobutton.label", TTK_PACK_LEFT)))
TTK_END_LAYOUT

TTK_BEGIN_LAYOUT(CheckbuttonLayout)
    TTK_NODE("Checkbutton.background", TTK_FILL_BOTH)
    TTK_GROUP("Checkbutton.button", TTK_FILL_BOTH,
	TTK_GROUP("Checkbutton.padding", TTK_FILL_BOTH|TTK_EXPAND,
	    TTK_NODE("Checkbutton.label", TTK_PACK_LEFT)))
TTK_END_LAYOUT

TTK_BEGIN_LAYOUT(MenubuttonLayout)
    TTK_NODE("Menubutton.background", TTK_FILL_BOTH)
    TTK_GROUP("Menubutton.button", TTK_FILL_BOTH,
	TTK_GROUP("Menubutton.padding", TTK_FILL_BOTH,
	    TTK_NODE("Menubutton.label", TTK_FILL_BOTH)))
TTK_END_LAYOUT

/*----------------------------------------------------------------------
 * +++ Initialization.
 */

int AquaTheme_Init(Tcl_Interp *interp)
{
    TTK_Theme themePtr, parentPtr;

    parentPtr = TTK_GetTheme(interp, "alt");
    themePtr = TTK_CreateTheme(interp, "aqua", parentPtr);
    if (!themePtr) {
        return TCL_ERROR;
    }

    /* Elements:
     */
    TTK_RegisterElementSpec(themePtr,"background",&BackgroundElementSpec,0);

    TTK_RegisterElementSpec(themePtr, "Button.button",
    	&ButtonElementSpec, ButtonParms + KTHEMEPushButton);
    TTK_RegisterElementSpec(themePtr, "Checkbutton.button",
    	&ButtonElementSpec, ButtonParms + KTHEMECheckBox);
    TTK_RegisterElementSpec(themePtr, "Radiobutton.button",
    	&ButtonElementSpec, ButtonParms + KTHEMERadioButton);
    TTK_RegisterElementSpec(themePtr, "Toolbutton.border",
    	&ButtonElementSpec, ButtonParms + KTHEMEBevelButton);
    TTK_RegisterElementSpec(themePtr, "Menubutton.button",
    	&ButtonElementSpec, ButtonParms + KTHEMEPopupButton);

    TTK_RegisterElementSpec(themePtr, "Notebook.tab", &TabElementSpec, 0);
    TTK_RegisterElementSpec(themePtr, "Notebook.client", &PaneElementSpec, 0);

    TTK_RegisterElementSpec(themePtr, "Labelframe.border",&GroupElementSpec,0);
    TTK_RegisterElementSpec(themePtr, "Entry.field",&EntryElementSpec,0);

    TTK_RegisterElementSpec(themePtr, "Combobox.field",&EntryElementSpec,0);
    TTK_RegisterElementSpec(themePtr, "Combobox.downarrow",
    	&PopupArrowElementSpec, 0);

    /* <<NOTE-TRACKS>>
     * The Progressbar widget adjusts the size of the bar element.
     * In the Aqua theme, the appearance manager computes the bar geometry,
     * so we make *.bar a no-op and do all the drawing in the *.trough
     */
    TTK_RegisterElementSpec(themePtr,"Progress.trough",
    	&TrackElementSpec, &ProgressData);
    TTK_RegisterElementSpec(themePtr,"Scale.trough",
    	&TrackElementSpec, &ScaleData);
    TTK_RegisterElementSpec(themePtr,"Progress.bar",&NullElementSpec,0);
    TTK_RegisterElementSpec(themePtr,"Scale.slider",&SliderElementSpec,0);

    /* Layouts:
     */
    TTK_RegisterLayout(themePtr, "TButton", ButtonLayout);
    TTK_RegisterLayout(themePtr, "TCheckbutton", CheckbuttonLayout);
    TTK_RegisterLayout(themePtr, "TRadiobutton", RadiobuttonLayout);
    TTK_RegisterLayout(themePtr, "TMenubutton", MenubuttonLayout);

    Tcl_PkgProvide(interp, "tile::theme::aqua", TILE_VERSION);
    return TCL_OK;
}

int TTK_MacPlatformInit(Tcl_Interp *interp)
{
    return AquaTheme_Init(interp);
}

