/*
 * xpTheme.c,v 1.60 2004/08/13 18:55:29 hobbs Exp
 *
 * Tk theme engine which uses the Windows XP "Visual Styles" API
 * Adapted from Georgios Petasis' XP theme patch.
 *
 * Copyright (c) 2003 by Georgios Petasis, petasis@iit.demokritos.gr.
 * Copyright (c) 2003 by Joe English
 * Copyright (c) 2003 by Pat Thoyts
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * See also:
 *
 * <URL: http://msdn.microsoft.com/library/en-us/
 *  	shellcc/platform/commctls/userex/refentry.asp >
 */

#ifdef _MSC_VER
#define WIN32_LEAN_AND_MEAN
#endif

#include <windows.h>

#ifdef NO_PRIVATE_HEADERS
#include <tk.h>
#include "gunk.h"
#else
#include <tkWinInt.h>
#endif

#include "tkTheme.h"

/*
 * This #if chunk makes it possible to build this library using older
 * versions of the Platform SDK that don't support the uxtheme library.
 */
#ifndef HAVE_UXTHEME_H

#if _MSC_VER > 100
#pragma message("warning: XP theme is unavailable.")
#endif
int XPTheme_Init(Tcl_Interp *interp) { return TCL_OK; }

#else

#include <uxtheme.h>
#include <tmschema.h>

typedef HTHEME  (STDAPICALLTYPE OpenThemeDataProc)(HWND hwnd,
		 LPCWSTR pszClassList);
typedef HRESULT (STDAPICALLTYPE CloseThemeDataProc)(HTHEME hTheme);
typedef HRESULT (STDAPICALLTYPE DrawThemeBackgroundProc)(HTHEME hTheme,
                 HDC hdc, int iPartId, int iStateId, const RECT *pRect,
                 OPTIONAL const RECT *pClipRect);
typedef HRESULT (STDAPICALLTYPE DrawThemeParentBackgroundProc)(HWND hwnd,
                 HDC hdc, OPTIONAL const RECT *prc);
typedef HRESULT (STDAPICALLTYPE DrawThemeEdgeProc)(HTHEME hTheme, HDC hdc,
		 int iPartId, int iStateId, const RECT *pDestRect,
		 UINT uEdge, UINT uFlags, RECT *pContentRect);
typedef HRESULT (STDAPICALLTYPE DrawThemeTextProc)(HTHEME hTheme, HDC hdc,
		 int iPartId, int iStateId, LPCWSTR pszText, int iCharCount,
		 DWORD dwTextFlags, DWORD dwTextFlags2, const RECT *pRect);
typedef HRESULT	(STDAPICALLTYPE GetThemeMarginsProc)(HTHEME, HDC,
		 int iPartId, int iStateId, int iPropId,
		 OPTIONAL RECT *prc, MARGINS *pMargins);
typedef HRESULT	(STDAPICALLTYPE GetThemePartSizeProc)(HTHEME,HDC,
		 int iPartId, int iStateId,
		 RECT *prc, enum THEMESIZE eSize, SIZE *psz);
typedef HRESULT (STDAPICALLTYPE GetThemeTextExtentProc)(HTHEME hTheme, HDC hdc,
		 int iPartId, int iStateId, LPCWSTR pszText, int iCharCount,
		 DWORD dwTextFlags, const RECT *pBoundingRect, RECT *pExtentRect);
typedef BOOL    (STDAPICALLTYPE IsThemeActiveProc)(VOID);

typedef struct
{
    OpenThemeDataProc			*OpenThemeData;
    CloseThemeDataProc			*CloseThemeData;
    DrawThemeBackgroundProc		*DrawThemeBackground;
    DrawThemeParentBackgroundProc	*DrawThemeParentBackground;
    DrawThemeEdgeProc			*DrawThemeEdge;
    DrawThemeTextProc		        *DrawThemeText;
    GetThemePartSizeProc		*GetThemePartSize;
    GetThemeTextExtentProc		*GetThemeTextExtent;
    IsThemeActiveProc                   *IsThemeActive;
} XPThemeProcs;

typedef struct
{
    HINSTANCE hlibrary;
    XPThemeProcs *procs;
} XPThemeData;

/*
 *----------------------------------------------------------------------
 *
 * LoadXPThemeProcs --
 *	Initialize XP theming support.
 *
 *	XP theme support is included in UXTHEME.DLL
 *	We dynamically load this DLL at runtime instead of linking
 *	to it at build-time.
 *
 * Returns:
 *	A pointer to an XPThemeProcs table if successful, NULL otherwise.
 */

static XPThemeProcs *
LoadXPThemeProcs(HINSTANCE *phlib)
{
    OSVERSIONINFO os;

    /*
     * We have to check whether we are running at least on Windows XP.
     * In order to determine this we call GetVersionEx directly, although
     * it would be a good idea to wrap it inside a function similar to
     * TkWinGetPlatformId...
     */
    ZeroMemory(&os, sizeof(os));
    os.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
    GetVersionEx(&os);
    if (os.dwMajorVersion >= 5 && os.dwMinorVersion >= 1) {
	/*
	 * We are running under Windows XP or a newer version.
	 * Load the library "uxtheme.dll", where the native widget
	 * drawing routines are implemented.
	 */
	HINSTANCE handle;
	*phlib = handle = LoadLibrary("uxtheme.dll");
	if (handle != 0)
	{
	    /*
	     * We have successfully loaded the library. Proceed in storing the
	     * addresses of the functions we want to use.
	     */
	    XPThemeProcs *procs = (XPThemeProcs*)ckalloc(sizeof(XPThemeProcs));
#define LOADPROC(name) \
	(0 != (procs->name = (name ## Proc *)GetProcAddress(handle, #name) ))

	    if (   LOADPROC(OpenThemeData)
		&& LOADPROC(CloseThemeData)
		&& LOADPROC(DrawThemeBackground)
		&& LOADPROC(DrawThemeParentBackground)
		&& LOADPROC(DrawThemeEdge)
		&& LOADPROC(DrawThemeText)
		&& LOADPROC(GetThemePartSize)
		&& LOADPROC(GetThemeTextExtent)
		&& LOADPROC(IsThemeActive)
	    )
	    {
		return procs;
	    }
#undef LOADPROC
	    ckfree((char*)procs);
	}
    }
    return 0;
}

/*
 * XPThemeDeleteProc --
 *
 *      Release any theme allocated resources.
 */

static void
XPThemeDeleteProc(void *clientData)
{
    XPThemeData *themeData = clientData;
    FreeLibrary(themeData->hlibrary);
    ckfree(clientData);
}

static int
XPThemeEnabled(TTK_Theme theme, void *clientData)
{
    XPThemeData *themeData = clientData;
    return themeData->procs->IsThemeActive();
}

/*
 * BoxToRect --
 * 	Helper routine.  Returns a RECT data structure.
 */
static RECT
BoxToRect(TTK_Box b)
{
    RECT rc;
    rc.top = b.y;
    rc.left = b.x;
    rc.bottom = b.y + b.height;
    rc.right = b.x + b.width;
    return rc;
}

/*
 * Map Tk state bitmaps to XP style enumerated values.
 */
static TTK_StateTable null_statemap[] = { {0,0,0} };

/*
 * Pushbuttons (Tk: "Button")
 */
static TTK_StateTable pushbutton_statemap[] =
{
    { PBS_DISABLED, 	TTK_STATE_DISABLED, 0 },
    { PBS_PRESSED, 	TTK_STATE_PRESSED, 0 },
    { PBS_HOT,		TTK_STATE_ACTIVE, 0 },
    { PBS_DEFAULTED,	TTK_STATE_DEFAULT, 0 },
    { PBS_NORMAL, 	0, 0 }
};

/*
 * Checkboxes (Tk: "Checkbutton")
 *
 * Missing: CBS_MIXEDDISABLED CBS_MIXEDHOT CBS_MIXEDNORMAL CBS_MIXEDPRESSED
 */
static TTK_StateTable checkbox_statemap[] =
{
{CBS_CHECKEDDISABLED,	TTK_STATE_DISABLED|TTK_STATE_SELECTED, 0},
{CBS_CHECKEDPRESSED,	TTK_STATE_PRESSED|TTK_STATE_SELECTED, 0},
{CBS_CHECKEDHOT,	TTK_STATE_ACTIVE|TTK_STATE_SELECTED, 0},
{CBS_CHECKEDNORMAL,	TTK_STATE_SELECTED, 0},
{CBS_UNCHECKEDDISABLED,	TTK_STATE_DISABLED, TTK_STATE_SELECTED},
{CBS_UNCHECKEDPRESSED,	TTK_STATE_PRESSED, TTK_STATE_SELECTED},
{CBS_UNCHECKEDHOT,	TTK_STATE_ACTIVE, TTK_STATE_SELECTED},
{CBS_UNCHECKEDNORMAL,	0,0 }
};

/*
 * Radiobuttons:
 */
static TTK_StateTable radiobutton_statemap[] =
{
{RBS_CHECKEDDISABLED,	TTK_STATE_DISABLED|TTK_STATE_SELECTED, 0},
{RBS_CHECKEDPRESSED,	TTK_STATE_PRESSED|TTK_STATE_SELECTED, 0},
{RBS_CHECKEDHOT,	TTK_STATE_ACTIVE|TTK_STATE_SELECTED, 0},
{RBS_CHECKEDNORMAL,	TTK_STATE_SELECTED, 0},
{RBS_UNCHECKEDDISABLED,	TTK_STATE_DISABLED, TTK_STATE_SELECTED},
{RBS_UNCHECKEDPRESSED,	TTK_STATE_PRESSED, TTK_STATE_SELECTED},
{RBS_UNCHECKEDHOT,	TTK_STATE_ACTIVE, TTK_STATE_SELECTED},
{RBS_UNCHECKEDNORMAL,	0,0 }
};

/*
 * Groupboxes (tk: "frame")
 */
static TTK_StateTable groupbox_statemap[] =
{
{GBS_DISABLED,	TTK_STATE_DISABLED, 0},
{GBS_NORMAL,	0,0 }
};

/*
 * Edit fields (tk: "entry")
 */

static TTK_StateTable edittext_statemap[] = 
{
    { ETS_DISABLED,	TTK_STATE_DISABLED, 0 },
    { ETS_READONLY,	TTK_STATE_READONLY, 0 },
    { ETS_FOCUSED,	TTK_STATE_FOCUS, 0 },
    { ETS_HOT,		TTK_STATE_ACTIVE, 0 },
    { ETS_NORMAL,	0, 0 }
/* NOT USED: ETS_ASSIST, ETS_SELECTED */
};

/*
 * Combobox button: (CBP_DROPDOWNBUTTON)
 */
static TTK_StateTable combobox_statemap[] = {
    { CBXS_DISABLED,	TTK_STATE_DISABLED, 0 },
    { CBXS_PRESSED, 	TTK_STATE_PRESSED, 0 },
    { CBXS_HOT, 	TTK_STATE_ACTIVE, 0 },
    { CBXS_NORMAL, 	0, 0 }
};

/*
 * Toolbar buttons (TP_BUTTON):
 */
static TTK_StateTable toolbutton_statemap[] =  {
    { TS_DISABLED, 	TTK_STATE_DISABLED, 0 },
    { TS_PRESSED,	TTK_STATE_PRESSED, 0 },
    { TS_HOTCHECKED,	TTK_STATE_SELECTED|TTK_STATE_ACTIVE, 0 },
    { TS_CHECKED, 	TTK_STATE_SELECTED, 0 },
    { TS_HOT,  		TTK_STATE_ACTIVE, 0 },
    { TS_NORMAL, 	0,0 }
};

/*
 * Scrollbars (Tk: "Scrollbar.thumb")
 */
static TTK_StateTable scrollbar_statemap[] =
{
    { SCRBS_DISABLED, 	TTK_STATE_DISABLED, 0 },
    { SCRBS_PRESSED, 	TTK_STATE_PRESSED, 0 },
    { SCRBS_HOT,	TTK_STATE_ACTIVE, 0 },
    { SCRBS_NORMAL, 	0, 0 }
};

static TTK_StateTable uparrow_statemap[] =
{
    { ABS_UPDISABLED,	TTK_STATE_DISABLED, 0 },
    { ABS_UPPRESSED, 	TTK_STATE_PRESSED, 0 },
    { ABS_UPHOT,	TTK_STATE_ACTIVE, 0 },
    { ABS_UPNORMAL, 	0, 0 }
};

static TTK_StateTable downarrow_statemap[] =
{
    { ABS_DOWNDISABLED,	TTK_STATE_DISABLED, 0 },
    { ABS_DOWNPRESSED, 	TTK_STATE_PRESSED, 0 },
    { ABS_DOWNHOT,	TTK_STATE_ACTIVE, 0 },
    { ABS_DOWNNORMAL, 	0, 0 }
};

static TTK_StateTable leftarrow_statemap[] =
{
    { ABS_LEFTDISABLED,	TTK_STATE_DISABLED, 0 },
    { ABS_LEFTPRESSED, 	TTK_STATE_PRESSED, 0 },
    { ABS_LEFTHOT,	TTK_STATE_ACTIVE, 0 },
    { ABS_LEFTNORMAL, 	0, 0 }
};

static TTK_StateTable rightarrow_statemap[] =
{
    { ABS_RIGHTDISABLED,TTK_STATE_DISABLED, 0 },
    { ABS_RIGHTPRESSED, TTK_STATE_PRESSED, 0 },
    { ABS_RIGHTHOT,	TTK_STATE_ACTIVE, 0 },
    { ABS_RIGHTNORMAL, 	0, 0 }
};

/*
 * Trackbar thumb: (Tk: "scale slider")
 */
static TTK_StateTable scale_statemap[] =
{
    { TUS_DISABLED, 	TTK_STATE_DISABLED, 0 },
    { TUS_PRESSED, 	TTK_STATE_PRESSED, 0 },
    { TUS_FOCUSED, 	TTK_STATE_FOCUS, 0 },
    { TUS_HOT,		TTK_STATE_ACTIVE, 0 },
    { TUS_NORMAL, 	0, 0 }
};

static TTK_StateTable tabitem_statemap[] =
{
    { TIS_DISABLED,     TTK_STATE_DISABLED, 0 },
    { TIS_SELECTED,     TTK_STATE_SELECTED, 0 },
    { TIS_HOT,          TTK_STATE_ACTIVE,   0 },
    { TIS_FOCUSED,      TTK_STATE_FOCUS,    0 },
    { TIS_NORMAL,       0,                  0 },
};

/*
 *----------------------------------------------------------------------
 * +++ Element data:
 *
 * The following structure is passed as the 'clientData' pointer
 * to most elements in this theme.  It contains data relevant
 * to a single XP Theme "part".
 *
 * <<NOTE-GetThemeMargins>>: 
 *	In theory, we should be call GetThemeMargins(...TMT_CONTENTRECT...)
 *	to calculate the internal padding.  In practice, this routine
 *	only seems to work properly for BP_PUSHBUTTON.  So we hardcode
 *	the required padding at element registration time instead.
 *
 * <<NOTE-GetThemePartSize>>: 
 *	This gives bogus metrics for some parts (in particular,
 *	BP_PUSHBUTTONS and 
 *	
 */
typedef struct
{
    /*
     * Static data, initialized when element is registered:
     */
    XPThemeProcs *procs;	/* Pointer to theme procedure table */
    LPCWSTR	className;	/* Windows window class name */
    int 	partId;		/* BP_PUSHBUTTON, BP_CHECKBUTTON, etc. */
    TTK_StateTable *statemap;	/* Map Tk states to XP states */
    TTK_Padding	padding;	/* See NOTE-GetThemeMargins */
    int  	flags;		
#   define 	IGNORE_THEMESIZE 0x1	/* See NOTE-GetThemePartSize */

    /*
     * Dynamic data, allocated by InitElementData:
     */
    HTHEME	hTheme;
    HDC		hDC;
    HWND	hwnd;

    /* For TkWinDrawableReleaseDC: */
    Drawable	drawable;
    TkWinDCState dcState;
} ElementData;

static ElementData *
NewElementData(
    XPThemeProcs *procs, LPCWSTR className, 
    int partId, TTK_StateTable *statemap, TTK_Padding pad, int flags)
{
    ElementData *elementData = (ElementData*)ckalloc(sizeof(ElementData));

    elementData->procs = procs;
    elementData->className = className;
    elementData->partId = partId;
    elementData->statemap = statemap;
    elementData->padding = pad;
    elementData->flags = flags;

    elementData->hTheme = elementData->hDC = 0;

    return elementData;
}

static void DestroyElementData(ElementData *elementData)
{
    ckfree((char*)elementData);
}

/*
 * InitElementData --
 * 	Looks up theme handle.  If Drawable argument is non-NULL,
 * 	also initializes DC.
 *
 * Returns:
 * 	1 on success, 0 on error.
 * 	Caller must later call FreeElementData() so this element
 * 	can be reused.
 */

static int
InitElementData(ElementData *elementData, Tk_Window tkwin, Drawable d)
{
    Window win = Tk_WindowId(tkwin);
    if (win == (Window)NULL)
	return 0;

    elementData->hwnd = Tk_GetHWND(win);
    elementData->hTheme = elementData->procs->OpenThemeData(
	elementData->hwnd, elementData->className);

    if (!elementData->hTheme)
	return 0;

    elementData->drawable = d;
    if (d != 0) {
	elementData->hDC = TkWinGetDrawableDC(Tk_Display(tkwin), d,
	    &elementData->dcState);
    }

    return 1;
}

static void
FreeElementData(ElementData *elementData)
{
    elementData->procs->CloseThemeData(elementData->hTheme);
    if (elementData->drawable != 0) {
	TkWinReleaseDrawableDC(
	    elementData->drawable, elementData->hDC, &elementData->dcState);
    }
}

/*----------------------------------------------------------------------
 * +++ Generic element implementation.
 * 
 * Used for elements which are handled entirely by the XP Theme API,
 * such as radiobutton and checkbutton indicators, scrollbar arrows, etc.
 */

static void
GenericElementGeometry(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, int *widthPtr, int *heightPtr, TTK_Padding *paddingPtr)
{
    ElementData *elementData = clientData;
    HRESULT result;
    SIZE size;

    if (!InitElementData(elementData, tkwin, 0))
	return;

    if (!(elementData->flags & IGNORE_THEMESIZE)) {
	result = elementData->procs->GetThemePartSize(
	    elementData->hTheme,
	    elementData->hDC,
	    elementData->partId,
	    TTK_StateTableLookup(elementData->statemap, 0),
	    NULL /*RECT *prc*/,
	    TS_TRUE,
	    &size);

	if (SUCCEEDED(result)) {
	    *widthPtr = size.cx;
	    *heightPtr = size.cy;
	} 
    }

    /* See NOTE-GetThemeMargins 
     */
    *paddingPtr = elementData->padding;

}

static void
GenericElementDraw(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, Drawable d,
    TTK_Box b, unsigned int state)
{
    ElementData *elementData = clientData;
    RECT rc = BoxToRect(b);

    if (!InitElementData(elementData, tkwin, d))
	return;

    elementData->procs->DrawThemeBackground(
	elementData->hTheme,
	elementData->hDC,
	elementData->partId,
	TTK_StateTableLookup(elementData->statemap, state),
	&rc,
	NULL/*pContentRect*/);

    FreeElementData(elementData);
}

static TTK_ElementSpec GenericElementSpec =
{
    TK_STYLE_VERSION_2,
    sizeof(NullElement),
    NullElementOptions,
    GenericElementGeometry,
    GenericElementDraw
};

/*----------------------------------------------------------------------
 * +++ Scrollbar thumb element.
 *     Same as a GenericElement, but also draws a grip.
 */

static void
ThumbElementDraw(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, Drawable d,
    TTK_Box b, unsigned int state)
{
    ElementData *elementData = clientData;
    unsigned stateId = TTK_StateTableLookup(elementData->statemap, state);
    RECT rc = BoxToRect(b);
    int gripperPartId;

    if (!InitElementData(elementData, tkwin, d))
	return;

    elementData->procs->DrawThemeBackground(elementData->hTheme,
	elementData->hDC, elementData->partId, stateId,
	&rc, NULL);

    /*
     * Draw grip:
     */
    if (elementData->partId == SBP_THUMBBTNVERT) {
	gripperPartId = SBP_GRIPPERVERT;
    } else {
	gripperPartId = SBP_GRIPPERHORZ;
    }

    elementData->procs->DrawThemeBackground(elementData->hTheme,
	elementData->hDC, gripperPartId,
	TTK_StateTableLookup(elementData->statemap, state),
	&rc, NULL);

    FreeElementData(elementData);
}

static TTK_ElementSpec ThumbElementSpec =
{
    TK_STYLE_VERSION_2,
    sizeof(NullElement),
    NullElementOptions,
    GenericElementGeometry,
    ThumbElementDraw
};


#if BROKEN_TEXT_ELEMENT

/*
 *----------------------------------------------------------------------
 * Text element (does not work yet).
 *
 * According to "Using Windows XP Visual Styles",  we need to select 
 * a font into the DC before calling DrawThemeText().
 * There's just no easy way to get an HFONT out of a Tk_Font.
 * Maybe GetThemeFont() would work?
 * 
 */

typedef struct
{
    Tcl_Obj *textObj;
    Tcl_Obj *fontObj;
} TextElement;

static TTK_ElementOptionSpec TextElementOptions[] =
{
    { "-text", TK_OPTION_STRING,
	Tk_Offset(TextElement,textObj), "" },
    { "-font", TK_OPTION_FONT,
	Tk_Offset(TextElement,fontObj), DEFAULT_FONT },
    { NULL }
};

static void
TextElementGeometry(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, int *widthPtr, int *heightPtr, TTK_Padding *paddingPtr)
{
    TextElement *element = elementRecord;
    ElementData *elementData = clientData;
    RECT rc = {0, 0};
    HRESULT hr = S_OK;

    if (!InitElementData(elementData, tkwin, 0))
	return;

    hr = elementData->procs->GetThemeTextExtent(
	    elementData->hTheme,
	    elementData->hDC,
	    elementData->partId,
	    TTK_StateTableLookup(elementData->statemap, 0),
	    Tcl_GetUnicode(element->textObj),
	    -1,
	    DT_LEFT,// | DT_BOTTOM | DT_NOPREFIX,
	    NULL,
	    &rc);

    if (SUCCEEDED(hr)) {
	*widthPtr = rc.right - rc.left;
	*heightPtr = rc.bottom - rc.top;
    }
    if (*widthPtr < 80) *widthPtr = 80;
    if (*heightPtr < 20) *heightPtr = 20;

    FreeElementData(elementData);
}

static void 
TextElementDraw(
    ClientData clientData, void *elementRecord,
    Tk_Window tkwin, Drawable d,
    TTK_Box b, unsigned int state)
{
    TextElement *element = elementRecord;
    ElementData *elementData = clientData;
    RECT rc = BoxToRect(b);
    HRESULT hr = S_OK;

    if (!InitElementData(elementData, tkwin, d))
	return;

    hr = elementData->procs->DrawThemeText(
	    elementData->hTheme,
	    elementData->hDC,
	    elementData->partId,
	    TTK_StateTableLookup(elementData->statemap, state),
	    Tcl_GetUnicode(element->textObj),
	    -1,
	    DT_LEFT,// | DT_BOTTOM | DT_NOPREFIX,
	    (state & TTK_STATE_DISABLED) ? DTT_GRAYED : 0,
	    &rc);
    FreeElementData(elementData);
}

static TTK_ElementSpec TextElementSpec =
{
    TK_STYLE_VERSION_2,
    sizeof(TextElement),
    TextElementOptions,
    TextElementGeometry,
    TextElementDraw
};

#endif	/* BROKEN_TEXT_ELEMENT */

/*----------------------------------------------------------------------
 * +++ Widget layouts:
 */

TTK_BEGIN_LAYOUT(ButtonLayout)
    TTK_NODE("Button.background", TTK_FILL_BOTH)
    TTK_GROUP("Button.button", TTK_FILL_BOTH,
	TTK_GROUP("Button.focus", TTK_FILL_BOTH, 
	    TTK_GROUP("Button.padding", TTK_FILL_BOTH,
		TTK_NODE("Button.label", TTK_FILL_BOTH))))
TTK_END_LAYOUT

TTK_BEGIN_LAYOUT(MenubuttonLayout)
    TTK_NODE("Menubutton.background", TTK_FILL_BOTH)
    TTK_NODE("Menubutton.dropdown", TTK_PACK_RIGHT|TTK_FILL_Y)
    TTK_GROUP("Menubutton.button", TTK_PACK_RIGHT|TTK_EXPAND|TTK_FILL_BOTH,
	    TTK_GROUP("Menubutton.padding", TTK_PACK_LEFT|TTK_EXPAND|TTK_FILL_X,
	        TTK_NODE("Menubutton.label", 0)))
TTK_END_LAYOUT

TTK_BEGIN_LAYOUT(VerticalScaleLayout)
    TTK_NODE("Scale.background", TTK_FILL_BOTH)
    TTK_GROUP("Scale.focus", TTK_EXPAND|TTK_FILL_BOTH,
	TTK_GROUP("Vertical.Scale.trough", TTK_EXPAND|TTK_FILL_BOTH,
	    TTK_NODE("Vertical.Scale.track", TTK_FILL_Y)
	    TTK_NODE("Vertical.Scale.slider", TTK_PACK_TOP) ))
TTK_END_LAYOUT

TTK_BEGIN_LAYOUT(HorizontalScaleLayout)
    TTK_NODE("Scale.background", TTK_FILL_BOTH)
    TTK_GROUP("Scale.focus", TTK_EXPAND|TTK_FILL_BOTH,
	TTK_GROUP("Horizontal.Scale.trough", TTK_EXPAND|TTK_FILL_BOTH,
	    TTK_NODE("Horizontal.Scale.track", TTK_FILL_X) 
	    TTK_NODE("Horizontal.Scale.slider", TTK_PACK_LEFT) ))
TTK_END_LAYOUT


/*----------------------------------------------------------------------
 * +++ Initialization routine:
 */

int XPTheme_Init(Tcl_Interp *interp)
{
    XPThemeData *themeData;
    XPThemeProcs *procs;
    HINSTANCE hlibrary;
    TTK_Theme themePtr, parentPtr;
    TTK_Padding nullPadding = TTK_UniformPadding(0);

    procs = LoadXPThemeProcs(&hlibrary);
    if (!procs)
	return TCL_ERROR;

    /*
     * Create the new style engine.
     */
    parentPtr = TTK_GetTheme(interp, "winnative");
    themePtr = TTK_CreateTheme(interp, "xpnative", parentPtr);

    if (!themePtr)
        return TCL_ERROR;

    /*
     * Set theme data and cleanup proc
     */

    themeData = (XPThemeData *)ckalloc(sizeof(XPThemeData));
    themeData->procs = procs;
    themeData->hlibrary = hlibrary;

    TTK_SetThemeEnabledProc(themePtr, XPThemeEnabled, themeData);
    TTK_RegisterCleanup(interp, themeData, XPThemeDeleteProc);

    /*
     * Register any new elements required for this style engine.
     */
#define GE(name,class,part,statemap) \
    TTK_RegisterElementSpec(themePtr, name, &GenericElementSpec, \
	NewElementData(procs, class, part, statemap, nullPadding, 0));

#define PAD TTK_MakePadding
#define GEPAD(name,class,part,statemap,padding, flags) \
    TTK_RegisterElementSpec(themePtr, name, &GenericElementSpec, \
	NewElementData(procs, class, part, statemap, padding, flags));

    TTK_RegisterElementSpec(themePtr, "Button.button", &GenericElementSpec,
	NewElementData(procs, L"BUTTON", BP_PUSHBUTTON, pushbutton_statemap,
	    PAD(3,3,3,3), IGNORE_THEMESIZE));

    /* Indicator elements 
     */
    GE( "Checkbutton.indicator",L"BUTTON",BP_CHECKBOX,checkbox_statemap);
    GE( "Radiobutton.indicator",L"BUTTON",BP_RADIOBUTTON,radiobutton_statemap);

    /* Frame elements */
    GEPAD("Labelframe.border", L"BUTTON", BP_GROUPBOX,
    	groupbox_statemap, PAD(2,2,2,2), 0);

    /* Entry border */
    GEPAD("Entry.field", L"EDIT", EP_EDITTEXT,
    	edittext_statemap, PAD(1,1,1,1), 0);
    GEPAD("Combobox.field", L"EDIT", EP_EDITTEXT,
    	edittext_statemap, PAD(1,1,1,1), 0);
    GE( "Combobox.downarrow", L"COMBOBOX",CP_DROPDOWNBUTTON,combobox_statemap);

#if BROKEN_TEXT_ELEMENT
    clientData = NewElementData(procs,
	L"BUTTON", BP_GROUPBOX, groupbox_statemap, 0);
    TTK_RegisterElementSpec(themePtr,
    	"Labelframe.text", &TextElementSpec, clientData, 0);
#endif /* BROKEN_TEXT_ELEMENT */

    /* Scrollbar elements */
    GE( "Vertical.Scrollbar.trough", 
    	L"SCROLLBAR", SBP_UPPERTRACKVERT, scrollbar_statemap);
    GE( "Horizontal.Scrollbar.trough", 
    	L"SCROLLBAR", SBP_UPPERTRACKHORZ, scrollbar_statemap);
    TTK_RegisterElementSpec(themePtr,
       "Horizontal.Scrollbar.thumb", &ThumbElementSpec,
	NewElementData(procs, L"SCROLLBAR", SBP_THUMBBTNHORZ, 
		scrollbar_statemap, nullPadding, 0));
    TTK_RegisterElementSpec(themePtr,
       "Vertical.Scrollbar.thumb", &ThumbElementSpec,
	NewElementData(procs,L"SCROLLBAR",SBP_THUMBBTNVERT,
	    scrollbar_statemap, nullPadding, 0));

    /* Arrow elements */
    GE( "Scrollbar.uparrow", L"SCROLLBAR",SBP_ARROWBTN,uparrow_statemap);
    GE( "Scrollbar.downarrow", L"SCROLLBAR",SBP_ARROWBTN,downarrow_statemap);
    GE( "Scrollbar.leftarrow", L"SCROLLBAR",SBP_ARROWBTN,leftarrow_statemap);
    GE( "Scrollbar.rightarrow", L"SCROLLBAR",SBP_ARROWBTN,rightarrow_statemap);

    /* Scale elements */
    GE( "Horizontal.Scale.slider",L"TRACKBAR", TKP_THUMBTOP, scale_statemap);
    GE( "Vertical.Scale.slider", L"TRACKBAR", TKP_THUMBRIGHT, scale_statemap);

    TTK_RegisterElementSpec(themePtr, "Scale.trough", &NullElementSpec, 0);
    GE( "Horizontal.Scale.track", L"TRACKBAR", TKP_TRACK, scale_statemap);
    GE( "Vertical.Scale.track", L"TRACKBAR", TKP_TRACKVERT, scale_statemap);

    /* Progress elements */
    GEPAD( "Horizontal.Progress.trough", L"PROGRESS", PP_BAR,
    	null_statemap, PAD(3,3,3,3), IGNORE_THEMESIZE);
    GEPAD( "Vertical.Progress.trough", L"PROGRESS", PP_BARVERT,
    	null_statemap, PAD(3,3,3,3), IGNORE_THEMESIZE);
    GE( "Horizontal.Progress.bar", L"PROGRESS", PP_CHUNK, null_statemap);
    GE( "Vertical.Progress.bar", L"PROGRESS", PP_CHUNKVERT, null_statemap);

    /* Notebook elements */
    GE( "Notebook.background", L"TAB", TABP_BODY, null_statemap);
    GE( "tab", L"TAB", TABP_TABITEM, tabitem_statemap);
    GEPAD( "client", L"TAB", TABP_PANE, null_statemap, PAD(1,1,3,3), 0);

    /* Toolbuttons: */
    GE( "Toolbutton.border", L"TOOLBAR", TP_BUTTON, toolbutton_statemap);
    GE( "Menubutton.button", L"TOOLBAR",
    	TP_SPLITBUTTON,toolbutton_statemap);
    GE( "Menubutton.dropdown", L"TOOLBAR",
    	TP_SPLITBUTTONDROPDOWN,toolbutton_statemap);

    /*
     * Layouts:
     */
    TTK_RegisterLayout(themePtr, "TButton", ButtonLayout);
    TTK_RegisterLayout(themePtr, "TMenubutton", MenubuttonLayout);
    TTK_RegisterLayout(themePtr, "Vertical.TScale", VerticalScaleLayout);
    TTK_RegisterLayout(themePtr, "Horizontal.TScale", HorizontalScaleLayout);

    Tcl_PkgProvide(interp, "tile::theme::xpnative", TILE_VERSION);

    return TCL_OK;
}

#endif /* WINVER < 0x0501 */
